#
# FtpCube
# Copyright (C) 2002 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import main
import dialog
import icons.folder

import os
from wxPython.wx import *

class Browser (dialog.DialogWindow):

	idTREE = wxNewId ()

	def __init__ (self, parent, title):
		dialog.DialogWindow.__init__ (self, parent, title)
		self.SetIcon (main.getApp ().getAppIcon ())

		panel = self.getDialogPanel ()

		label = wxStaticText (panel, -1, _("Directory:"))
		self.dir_entry = wxTextCtrl (panel, -1)
		dsizer = wxFlexGridSizer (5, 3)
		dsizer.AddMany ([
			(20, 5), (label, 0, wxEXPAND), (20, 5),
			(20, 5), (20, 5), (20, 5),
			(20, 5), (self.dir_entry, 1, wxEXPAND), (20, 5),
		])
		dsizer.AddGrowableCol (1)

		self.tree = wxTreeCtrl (panel, self.idTREE, style=wxSUNKEN_BORDER | wxTR_HAS_BUTTONS)
		self.image_list = wxImageList (15, 17)
		self.folder_index = self.image_list.Add (icons.folder.getBitmap ())
		self.tree.SetImageList (self.image_list)
		tsizer = wxBoxSizer (wxHORIZONTAL)
		tsizer.Add (15, 5)
		tsizer.Add (self.tree, 1, wxEXPAND)
		tsizer.Add (10, 5)

		sizer = wxBoxSizer (wxVERTICAL)
		sizer.Add (5, 5)
		sizer.Add (dsizer, 0, wxEXPAND)
		sizer.Add (5, 10)
		sizer.Add (tsizer, 1, wxEXPAND)
		sizer.Add (5, 5)
		panel.SetAutoLayout (true)
		panel.SetSizer (sizer)
		sizer.Fit (panel)
		sizer.SetSizeHints (panel)

		# Render the dialog buttons and set the sizers
		self.SetSizeHints (400, 350)
		self.renderDialog ()

		EVT_TREE_SEL_CHANGED (self, self.idTREE, self.onTreeSelect)

	def getDirectory (self):
		return self.dir_entry.GetValue ()

	def onTreeSelect (self, event):
		raise NotImplementedError

class LocalBrowser (Browser):

	def __init__ (self, parent, path):
		Browser.__init__ (self, parent, _("Ftpcube - Change Local Directory"))

		self.root = self.tree.AddRoot (os.sep, self.folder_index, self.folder_index)
		self.loadTree (path, self.root)

		# Set the entry box
		self.dir_entry.SetValue (path)

	def loadTree (self, path, parent):
		if not os.path.isdir (path):
			path = os.path.dirname (path)

		# Create a list of paths to cycle through
		dirs = path.split (os.sep)
		dirs[0] = os.sep

		cur_path = os.sep
		parent_list = [ parent ]
		while dirs:
			dir = dirs.pop (0)
			cur_path = os.path.join (cur_path, dir)
			listing = self.getDirectoryList (cur_path)

			new_parent = None
			for item in listing:
				node = self.tree.AppendItem (parent, item, self.folder_index,
				                                           self.folder_index)
				if dirs and item == dirs[0]:
					new_parent = node
			if new_parent:
				parent = new_parent
				parent_list.append (parent)

		# Expand all parents
		for p in parent_list:
			self.tree.Expand (p)

	def getDirectoryList (self, path):
		try:
			listing = os.listdir (path)
		except OSError, strerror:
			return [ ]
		# Filter so we're only looking at directories
		listing = [ x for x in listing
		              if os.path.isdir (os.path.join (path, x)) ]
		return listing

	def onTreeSelect (self, event):
		node = event.GetItem ()
		path = self.nodeToPath (node)
		self.dir_entry.SetValue (path)

		if not self.tree.GetChildrenCount (node):
			self.expandNode (node, path)
			self.tree.Expand (node)

	def expandNode (self, parent, path):
		listing = self.getDirectoryList (path)
		self.tree.Freeze ()
		for item in listing:
			node = self.tree.AppendItem (parent, item, self.folder_index,
			                                           self.folder_index)
		self.tree.Thaw ()

	def nodeToPath (self, node):
		path = [ ]
		while node != self.root:
			path.insert (0, self.tree.GetItemText (node))
			node = self.tree.GetItemParent (node)
		path = os.sep + os.sep.join (path)
		return path

class RemoteBrowser (Browser):

	def __init__ (self, parent, cache, path):
		Browser.__init__ (self, parent, _("Ftpcube - Change Remote Directory"))
		self.cache = cache

		# Load in the tree contents
		self.root = self.tree.AddRoot (self.cache.getDelimiter (), self.folder_index,
		                                                           self.folder_index)
		self.loadTree (self.cache.getRoot (), self.root)
		self.expandPath (path)

		# Set the entry box
		self.dir_entry.SetValue (path)

	def onTreeSelect (self, event):
		node = event.GetItem ()
		path = self.nodeToPath (node)
		self.dir_entry.SetValue (path)

		if not self.tree.GetChildrenCount (node):
			self.expandPath (path)

	def loadTree (self, node, parent):
		directories = [ x for x in node.data
		                if not x[0] == '..' and x[3][0] == 'd' ]
		parent_list = [ ]
		for dir in directories:
			dir_name = dir[0]
			insert_node = self.tree.AppendItem (parent, dir_name, self.folder_index,
			                                                      self.folder_index)
			child = self.cache.getChildNode (node, dir_name)
			if child is not None:
				self.loadTree (child, insert_node)

	def nodeToPath (self, node):
		path = [ ]
		while node != self.root:
			path.insert (0, self.tree.GetItemText (node))
			node = self.tree.GetItemParent (node)
		path = self.tree.GetItemText (self.root) + self.cache.getDelimiter ().join (path)
		return path

	def expandPath (self, path):
		path = [ x for x in path.split (self.cache.getDelimiter ()) if x ]
		self.tree.Expand (self.root)
		if not path:
			return

		node = self.root
		while path:
			node_name = path.pop (0)

			# Extract the children from the tree
			children = [ ]
			child, cookie = self.tree.GetFirstChild (node, 0)
			last_child = self.tree.GetLastChild (node)
			children.append (child)
			while child != last_child:
				child, cookie = self.tree.GetNextChild (node, cookie)
				children.append (child)
			children.append (last_child)

			# Expand the appropriate child
			child_names = [ self.tree.GetItemText (x) for x in children ]
			try:
				child_index = child_names.index (node_name)
			except ValueError:
				return None
			node = children[child_index]
			self.tree.Expand (node)
