#
# FtpCube
# Copyright (C) 2002 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import logger
import threading
import time
import exceptions

class DispatcherException (exceptions.Exception):

	def __init__ (self, args=None):
		self.args = args

class Dispatcher (threading.Thread):

	def __init__ (self, **kwargs):
		threading.Thread.__init__ (self)
		self.setDaemon (1)

		# Try grabbing the logger argument
		try:
			self.logger = kwargs['logger']
		except KeyError:
			self.logger = None
		if self.logger and not isinstance (self.logger, logger.Logger):
			raise DispatcherException, _("Invalid logger: %(log)s") \
			                             %{ 'log' : self.logger}

		# Initialize instance variables
		self.cmdobj = None
		self.die = False
		self.response = None
		self.outgoing_event = threading.Event ()
		self.outgoing = [ ]
		self.default_handler = None

	def setDefaultHandler (self, handler):
		self.default_handler = handler

	def getDefaultHandler (self):
		return self.default_handler

	def resetQueue (self):
		self.outgoing = [ ]
		self.outgoing_event.set ()

	def destroy (self):
		self.die = True
		self.resetQueue ()

	def run (self):
		while not self.die:
			if not self.outgoing:
				self.outgoing_event.clear ()
				self.outgoing_event.wait ()

			if self.outgoing:
				self.cmdobj = self.outgoing.pop (0)

				if __debug__:
					print "%s dispatching: %s" %(threading.currentThread (), self.cmdobj)

				if not isinstance (self.cmdobj, DispatcherCmd):
					self.log ('error',
					          _("Warning: Attempting to dispatch invalid command object: %(obj)s")
					          %{ 'obj' : self.cmdobj})

				try:
					# Dispatch the command object in our thread
					self.response = self.cmdobj.execute (self)
				except Exception, strerror:
					if __debug__:
						print "Dispatcher caught error executing command: %s: %s" %(Exception, strerror)
					if not self.die:
						handler = self.cmdobj.getErrorHandler ()
						if isinstance (handler, ErrorHandler):
							handler.handle (strerror, self)
						elif isinstance (self.default_handler, ErrorHandler):
							self.default_handler.handle (strerror, self)
						else:
							self.log ('error', strerror)
		if __debug__:
			print "Terminating Thread: %s" %repr (self)

	def dispatch (self, cmdobj):
		self.outgoing.append (cmdobj)
		if not self.outgoing_event.isSet ():
			self.outgoing_event.set ()

	def getExecutingCmd (self):
		return self.cmdobj

	def busy (self):
		return self.outgoing_event.isSet ()

	def log (self, kind, msg):
		if self.logger:
			self.logger.log (kind, msg)

class DispatcherCmd:

	def __init__ (self):
		self.error_handler = None

	def setErrorHandler (self, handler):
		self.error_handler = handler

	def getErrorHandler (self):
		return self.error_handler

	def execute (self, dispatcher):
		raise NotImplementedError, _("execute method must be implemented")

class DispatchStatus:

	def __init__ (self):
		self.opt_data = None

	def setOptionalData (self, data):
		self.opt_data = data

	def getOptionalData (self):
		return self.opt_data

	def start (self):
		raise NotImplementedError, _("start method must be implemented")

	def update (self, update):
		raise NotImplementedError, _("update method must be implemented")

	def finished (self):
		raise NotImplementedError, _("finished method must be implemented")

class ErrorHandler:

	def __init__ (self):
		pass

	def handle (self, error, dispatcher):
		# Do nothing by default
		pass

class ErrorLogger (ErrorHandler):

	def __init__ (self, mylogger):
		ErrorHandler.__init__ (self)
		if not isinstance (mylogger, logger.Logger):
			raise TypeError, _("expected Logger type")
		self.logger = mylogger

	def handle (self, error, dispatcher):
		self.logger.log ('error', error)
