#
# FtpCube
# Copyright (C) 2002 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import ftp
import dispatcher
import browsewin
import ctrlwin
import messages
import events
import protocol
import main

import os
import time
import exceptions
from wxPython.wx import *

class ThreadException (exceptions.Exception):

	def __init__ (self, args=None):
		self.args = args

class ThreadManager:

	def __init__ (self):
		self.threads = [ None ]

		# Listeners for thread manager events
		self.listeners = [ ]

	def getThreadList (self):
		return self.threads

	def getThread (self, id):
		try:
			return self.threads[id]
		except IndexError:
			return None

	def newMainThread (self, **kwargs):
		self.threads[0] = BrowseThread (**kwargs)
		self.threads[0].setId (0)
		event = events.wxThreadEvent (0, events.wxThreadEvent.EVT_CREATE)
		self.notifyListeners (event)
		return 0

	def getMainThread (self):
		return self.threads[0]

	def removeMainThread (self):
		self.removeThread (0)

	def newTransferThread (self, **kwargs):
		thread = TransferThread (**kwargs)
		id = len (self.threads)
		thread.setId (id)
		self.threads.append (thread)
		if __debug__:
			print "Creating new transfer thread and setting id to: [%s]" %id
		event = events.wxThreadEvent (id, events.wxThreadEvent.EVT_CREATE)
		self.notifyListeners (event)
		return id

	def removeTransferThread (self, id):
		self.removeThread (id)

	def removeThread (self, id):
		try:
			self.threads[id] = None
		except:
			# Do nothing if we were given a bad ID
			pass
		event = events.wxThreadEvent (id, events.wxThreadEvent.EVT_DESTROY)
		self.notifyListeners (event)

	def addListener (self, win):
		self.listeners.append (win)

	def removeListener (self, win):
		self.listeners.remove (win)

	def notifyListeners (self, event):
		for each in self.listeners:
			wxPostEvent (each, event)

class ConnectionThread:

	def __init__ (self, **kwargs):
		# Check if this thread has a name
		try:
			self.name = kwargs['name']
		except KeyError:
			self.name = None

		# Fetch the connection options
		try:
			if not kwargs.has_key ('opts'):
				raise Exception, _("Missing connection options field")
			self.opts = kwargs['opts']
			if not isinstance (self.opts, dict):
				raise Exception, _("Invalid connection options")
		except Exception, strerror:
			raise ThreadException, strerror

		# Grab our logging mechanisms
		try:
			self.logger = kwargs['logger']
		except KeyError:
			self.logger = None
		try:
			self.download_logger = kwargs['download_logger']
		except KeyError:
			self.download_logger = None
		try:
			self.upload_logger = kwargs['upload_logger']
		except KeyError:
			self.upload_logger = None

		# Initialize instance variables
		self.id = None
		self.progress = (0, 0)
		self.start_time = None
		self.idle = 0
		self.idle_enable = True
		self.done = False
		self.connect = None
		self.transfer_state = None

	def setId (self, id):
		self.id = id

	def getId (self):
		return self.id

	def getName (self):
		return self.name

	def postEvent (self, kind, data=None):
		event = events.wxThreadEvent (self.id, kind, data=data)
		thread_mgr = main.getApp ().getThreadManager ()
		thread_mgr.notifyListeners (event)

	def initiateConnect (self):
		self.postEvent (events.wxThreadEvent.EVT_CONNECT)

		try:
			self.connect = ftp.FTP (
				host=self.opts['host'],
				port=self.opts['port'],
				logger=self.logger,
			)
		except ftp.FtpException, strerror:
			raise ThreadException, strerror

		self.connect.initiateConnect (handler=self.getConnectHandler ())
		self.connect.getWelcomeMsg (handler=self.getConnectHandler ())
		self.connect.login (self.opts['username'], self.opts['password'],
		                    handler=self.getConnectHandler ())

	def getConnectHandler (self):
		handler = ConnectHandler (self, self.logger)
		return handler

	def cancelConnect (self):
		if self.connect:
			self.connect.destroy ()

	def finished (self):
		return self.done

	def setDone (self):
		self.done = True

	def isTransfering (self):
		cur, total = self.progress
		return not (cur == 0 or cur == total)

	def getConnection (self):
		return self.connect

	def getOpts (self):
		opts = { }
		opts.update (self.opts)
		return opts

	def getHost (self):
		return self.opts['host']

	def getPort (self):
		return self.opts['port']

	def getPassive (self):
		return self.opts['passive']

	def getRetries (self):
		return self.opts['retries']

	def getProgress (self):
		return self.progress

	def isIdleEnabled (self):
		return self.idle_enable

	def setIdleEnable (self, bool):
		self.idle_enable = bool

	def getIdleTime (self):
		return self.idle

	def updateIdleTime (self):
		self.idle = self.idle + 1

		# Check if we should send a NOOP to prevent idling
		if (self.idle % self.opts['timeout']) == 0:
			self.connect.noop ()

	def getTransferSpeed (self):
		elapsed = time.time () - self.start_time
		if elapsed:
			return (float (self.progress[0]) / float (elapsed)) / 1024 # In KB
		return 0

	def updateProgress (self, cur, max):
		self.progress = (cur, max)

	def saveTransferState (self, event):
		if not isinstance (event, events.TransferEvent):
			raise TypeError, _("Expected Queue Event")
		self.transfer_state = event

	def loadTransferState (self):
		return self.transfer_state

	def getLocalFileSize (self, path):
		file_size = None
		# Check if the file exists and whether or not we should resume
		if os.path.exists (path) and os.path.isfile (path):
			try:
				file_size = os.path.getsize (path)
			except OSError, strerror:
				messages.displayErrorDialog (main.getApp ().getMainWindow (),
					_("Couldn't obtain file size: %(err)s")
					%{ 'err' : strerror })
				raise
		return file_size

	def initiateTransfer (self, event, handler=None):
		self.saveTransferState (event)
		self.cwd (event.remote_path)
		self.start_time = time.time ()
		self.progress = (0, event.size)

		# For whether we need resuming
		rest = None
		file_size = None

		local_file = os.path.join (event.local_path, event.file)
		try:
			file_size = self.getLocalFileSize (local_file)
		except:
			return

		if event.direction == protocol.DOWNLOAD:

			# Skip the file if it's exactly the same size. This means that a user
			# must delete a rogue file if the sizes match
			if file_size == event.size:
				return

			if file_size is not None and file_size < event.size:
				rest = file_size
				self.progress = (file_size, event.size)

			# Dispatcher status class for transfer
			status = TransferStatus (self, event, self.progress[0], self.progress[1])
			if event.method == protocol.BINARY:
				self.connect.retrieveBinary (event.file, event.local_path,
				                             status=status,
				                             rest=rest,
				                             passive=self.getPassive (),
				                             handler=handler)
			else:
				self.connect.retrieveAscii (event.file, event.local_path,
				                            status=status,
				                            passive=self.getPassive (),
				                            handler=handler)
			self.logTransfer ('download',
			                  event.username,
			                  event.host,
			                  '/'.join ([ event.remote_path, event.file ]),
			                  os.path.join ([ event.local_path, event.file ]),
			                  event.size)
		elif event.direction == protocol.UPLOAD:

			if file_size is not None and event.size < file_size:
				rest = event.size
				self.progress = (event.size, file_size)

			# Dispatcher status class for transfer
			status = TransferStatus (self, event, self.progress[0], self.progress[1])
			if event.method == protocol.BINARY:
				self.connect.uploadBinary (event.file, event.local_path,
				                           status=status, rest=rest, handler=handler)
			else:
				self.connect.uploadAscii (event.file, event.local_path,
				                          status=status, handler=handler)
			self.logTransfer ('upload',
			                  event.username,
			                  event.host,
			                  os.path.join (event.local_path, event.file),
			                  '/'.join ([ event.remote_path, event.file ]),
			                  event.size)
		else:
			raise TypeError, "Invalid direction for ftp transfer: [%s]" %event.direction

	def finishTransferEvent (self, event):
		if event.direction == protocol.DOWNLOAD:
			# Grab the file size so we can check if our transfer was
			# completed. If not, it'll go into the failure window
			# as an incomplete.
			try:
				file_size = self.getLocalFileSize (os.path.join (event.local_path, event.file))
			except:
				file_size = None

			if file_size != event.size:
				opts = event.toHash ()
				# Explicitly set the attempts to max attempts so this will appear in the
				# failure window
				opts['attempt'] = opts['maxattempts']
				opts.update ({ 'error' : _("Transfer incomplete - File sizes do not match") })
				fail_evt = events.wxFailureEvent (**opts)
				main.getApp ().postEvent (fail_evt)
			else:
				# Recast event as download event
				down_evt = events.wxDownloadEvent (**event.toHash ())
				main.getApp ().postEvent (down_evt)

			# Refresh the local window if needed
			local_win = main.getApp ().getMainWindow ().getLocalWindow ()
			if local_win.getDir () == event.local_path:
				win_evt = events.wxLocalWindowEvent (
					events.wxLocalWindowEvent.EVT_LIST_REFRESH,
					event.remote_path,
				)
				main.getApp ().postEvent (win_evt)
		elif event.direction == protocol.UPLOAD:
			event = events.wxRemoteWindowEvent (
				events.wxRemoteWindowEvent.EVT_LIST_REFRESH,
				event.remote_path,
			)
			main.getApp ().postEvent (event)

	def logTransfer (self, direction, user, host, src, dest, size):
		msg = _("%s@%s: %s -> %s [%d bytes]\n") %(user, host, src, dest, size)
		if direction == 'download' and self.download_logger:
			self.download_logger.log (msg)
		elif direction == 'upload' and self.upload_logger:
			self.upload_logger.log (msg)

	def recurseDirectory (self, event, handler=None):
		self.saveTransferState (event)
		self.postEvent (events.wxThreadEvent.EVT_LIST)

		if event.remote_path == '/':
			remote_path = event.remote_path + event.file
		else:
			remote_path = '/'.join ([ event.remote_path, event.file ])

		local_path = os.path.join (event.local_path, event.file)
		if event.direction == protocol.DOWNLOAD:
			if not os.path.exists (local_path):
				try:
					os.makedirs (local_path)
				except OSError, strerror:
					messages.displayErrorDialog (main.getApp ().getMainWindow (),
						_("Error creating local directories: ") + strerror)
					return

			self.connect.cwd (remote_path)
			event_copy = events.wxQueueEvent (**event.toHash ())
			event_copy.remote_path = remote_path
			event_copy.local_path  = local_path
			status = DirectoryStatus (self, event_copy)
			self.connect.list (status=status, passive=self.getPassive ())
		elif event.direction == protocol.UPLOAD:
			self.connect.mkdir (remote_path)
			try:
				files = os.listdir (local_path)
			except OSError, strerror:
				messages.displayErrorDialog (main.getApp ().getMainWindow (),
					 _("Error reading directory: ") + strerror)
				return
			for f in files:
				new_path = os.path.join (local_path, f)
				try:
					size = os.path.getsize (new_path)
				except OSError, strerror:
					messagtes.displayErrorDialog (main.getApp ().getMainWindow (),
						_("Error obtaining size for file %(file)s: %(err)s")
						%{ 'file' : f, 'err' : strerror })
					continue
				if os.path.isfile (new_path):
					flag = 'f'
				elif os.path.isdir (new_path):
					flag = 'd'
				elif os.path.islink (new_path):
					flag = 'l'
				else:
					flag = '-'
				new_event = events.wxQueueEvent (
					file        = f,
					host        = event.host,
					port        = event.port,
					username    = event.username,
					password    = event.password,
					size        = size,
					flags       = flag,
					remote_path = remote_path,
					local_path  = local_path,
					direction   = event.direction,
					attempt     = 1,
					maxattempts = event.maxattempts,
					method      = event.method,
				)
				main.getApp ().postEvent (new_event)
		else:
			raise TypeError, "Invalid direction for ftp transfer: [%s]" %event.direction

	def recurseLink (self, event):
		class Handler (dispatcher.ErrorHandler):
			def __init__ (self, thread):
				self.thread = thread
			def handle (self, error, dispatcher):
				event = self.thread.loadTransferState ()
				self.thread.recurseDirectory (event)

		handler = Handler (self)
		self.initiateTransfer (event, handler=handler)

	def noop (self, handler=None):
		self.idle = 0
		self.connect.noop (handler=handler)

	def cwd (self, dir, handler=None):
		self.postEvent (events.wxThreadEvent.EVT_CWD)
		self.idle = 0
		self.connect.cwd (dir, handler=handler)

	def list (self, handler=None):
		self.postEvent (events.wxThreadEvent.EVT_LIST)
		self.idle = 0
		list_handler = ListStatus ()
		self.connect.list (status=list_handler, passive=self.getPassive (), handler=handler)

	def pwd (self, status=None, handler=None):
		self.idle = 0
		return self.connect.pwd (status=status, handler=handler)

	def rename (self, old_name, new_name, handler=None):
		data = (old_name, new_name)
		self.postEvent (events.wxThreadEvent.EVT_RENAME, data=data)
		self.idle = 0
		self.connect.rename (old_name, new_name, handler=handler)

	def delete (self, name, handler=None):
		self.postEvent (events.wxThreadEvent.EVT_DELETE)
		self.idle = 0
		self.connect.delete (name, handler=handler)

	def mkdir (self, name, handler=None):
		self.postEvent (events.wxThreadEvent.EVT_MKDIR)
		self.idle = 0
		self.connect.mkdir (name, handler=handler)

	def rmdir (self, name, handler=None):
		self.postEvent (events.wxThreadEvent.EVT_RMDIR)
		self.idle = 0
		self.connect.rmdir (name, handler=handler)

	def chmod (self, name, perm, handler=None):
		self.postEvent (events.wxThreadEvent.EVT_CHMOD)
		self.idle = 0
		self.connect.chmod (perm, name, handler=handler)

	def quit (self):
		self.connect.quit ()

	def abort (self):
		if self.connect:
			self.connect.abort ()

	def busy (self):
		if self.connect is None:
			return False
		return self.connect.busy ()

class BrowseThread (ConnectionThread):

	def __init__ (self, **kwargs):
		ConnectionThread.__init__ (self, **kwargs)
		self.abort_connect = False

	def initiateConnect (self):
		self.abort_connect = False
		self.attempt = 0
		ConnectionThread.initiateConnect (self)
		self.completeConnection ()

	def completeConnection (self):
		# Change to appropriate local directory
		if self.opts['localdir']:
			evt = events.wxLocalWindowEvent (
				events.wxLocalWindowEvent.EVT_LIST_UPDATE,
				self.opts['localdir']
			)
			main.getApp ().postEvent (evt)

		# Change to appropriate remote directory
		if self.opts['remotedir']:
			self.connect.cwd (self.opts['remotedir'])

		# Make these calls directly so we don't update the
		# GUI status entry, since this is still considered
		# connecting.
		pwd_status = PwdStatus ()
		self.connect.pwd (status=pwd_status)
		list_handler = ListStatus ()
		self.connect.list (status=list_handler, passive=self.getPassive ())

	def retryConnection (self, attempt):
		self.attempt = attempt
		ConnectionThread.initiateConnect (self)
		# This will get wiped out of the queue if the handler is called
		self.completeConnection ()

	def abort (self):
		if self.abort_connect:
			ConnectionThread.abort (self)
		else:
			self.abort_connect = True

	def shouldAbortConnection (self):
		return self.abort_connect or self.finished ()

	def getConnectHandler (self):
		handler = BrowseConnectHandler (self, self.attempt, self.logger)
		return handler

	def updateIdleTime (self):
		ConnectionThread.updateIdleTime (self)

		# Check if it's time to give up and disconnect
		if self.idle == self.opts['main_idle']:
			main.getApp ().getMainWindow ().onDisconnect (None)

	def initiateTransfer (self, event):
		dispatcher = self.connect.getDispatcher ()
		remote_win = main.getApp ().getMainWindow ().getRemoteWindow ()
		old_dir = remote_win.getDir ()
		ConnectionThread.initiateTransfer (self, event)
		self.connect.cwd (old_dir)

	def recurseDirectory (self, event):
		remote_win = main.getApp ().getMainWindow ().getRemoteWindow ()
		old_dir = remote_win.getDir ()
		ConnectionThread.recurseDirectory (self, event)
		self.connect.cwd (old_dir)

	def recurseLink (self, event):
		remote_win = main.getApp ().getMainWindow ().getRemoteWindow ()
		old_dir = remote_win.getDir ()
		ConnectionThread.recurseLink (self, event)
		self.connect.cwd (old_dir)

class TransferThread (ConnectionThread):

	def initiateConnect (self):
		handler = DefaultErrorHandler (self)
		try:
			ConnectionThread.initiateConnect (self)
		except Exception, strerror:
			handler.handle (strerror, None)
		self.connect.getDispatcher ().setDefaultHandler (handler)

	def updateIdleTime (self):
		ConnectionThread.updateIdleTime (self)

		# Check if it's time to give up and disconnect
		if self.idle == self.opts['thread_idle']:
			self.setDone ()

class PwdStatus (dispatcher.DispatchStatus):

	def __init__ (self):
		dispatcher.DispatchStatus.__init__ (self)

	def start (self):
		pass

	def update (self, update):
		pass

	def finished (self):
		data = self.getOptionalData ()
		event = events.wxRemoteWindowEvent (
			events.wxRemoteWindowEvent.EVT_CWD,
			data,
		)
		main.getApp ().postEvent (event)


class ListStatus (dispatcher.DispatchStatus):

	def __init__ (self):
		dispatcher.DispatchStatus.__init__ (self)

	def start (self):
		pass

	def update (self, update):
		pass

	def finished (self):
		data = self.getOptionalData ()
		event = events.wxRemoteWindowEvent (
			events.wxRemoteWindowEvent.EVT_LIST_UPDATE,
			data
		)
		main.getApp ().postEvent (event)

class DefaultErrorHandler (dispatcher.ErrorHandler):

	def __init__ (self, thread):
		dispatcher.ErrorHandler.__init__ (self)
		self.thread = thread

	def handle (self, error, dispatcher):
		event = self.thread.loadTransferState ()
		if isinstance (event, events.TransferEvent):
			opts = event.toHash ()
			opts['error'] = error
			new_event = events.wxFailureEvent (**opts)
			main.getApp ().postEvent (new_event)

class ConnectHandler (DefaultErrorHandler):

	def __init__ (self, thread, logger):
		DefaultErrorHandler.__init__ (self, thread)
		self.logger = logger

	def handle (self, error, dispatcher):
		if self.logger:
			self.logger.log ('error', _("Thread Connection Error: %(err)s")
			                          %{ 'err' : error})
		DefaultErrorHandler.handle (self, error, dispatcher)
		self.thread.cancelConnect ()

class BrowseConnectHandler (dispatcher.ErrorHandler):

	def __init__ (self, thread, attempt, logger):
		dispatcher.ErrorHandler.__init__ (self)
		self.thread = thread
		self.opts = self.thread.getOpts ()
		self.connect = self.thread.getConnection ()
		self.attempt = attempt
		self.logger = logger

	def handle (self, error, dispatcher):
		if self.attempt >= self.opts['retries']:
			if self.logger:
				self.logger.log ('error',
					_("Maximum connection attempt hit. Aborting connection..."))
			self.destroyConnect ()
			return

		delay = self.opts['delay']
		if self.logger:
			self.logger.log ('error',
				_("Waiting %(delay)s secs before re-attempting connection...")
				%{ 'delay' : delay })

		slept = 0
		while slept < delay:
			time.sleep (1)
			slept = slept + 1

			# If we should abort, cancel the connection thread
			if self.thread.shouldAbortConnection ():
				self.destroyConnect ()
				break

		self.cancelConnect ()
		if not self.thread.shouldAbortConnection ():
			self.thread.retryConnection (self.attempt + 1)

	def cancelConnect (self):
		if self.connect is not None:
			self.connect.destroy ()
			self.thread.connect = None

	def destroyConnect (self):
		self.cancelConnect ()
		self.thread.setDone ()

class TransferStatus (dispatcher.DispatchStatus):

	def __init__ (self, thread, event, min, max):
		dispatcher.DispatchStatus.__init__ (self)
		self.thread = thread
		self.event = event
		self.progress = (min, max)

	def start (self):
		cur, max = self.progress
		self.thread.updateProgress (cur, max)

	def update (self, update):
		cur, max = self.progress
		cur = cur + update
		self.thread.updateProgress (cur, max)
		self.progress = (cur, max)

	def finished (self):
		self.thread.finishTransferEvent (self.event)

class DirectoryStatus (dispatcher.DispatchStatus):

	def __init__ (self, thread, event):
		dispatcher.DispatchStatus.__init__ (self)
		self.thread = thread
		self.event = event

	def start (self):
		pass

	def update (self, update):
		pass

	def finished (self):
		listing = self.getOptionalData ()
		files = [ ]
		remote_window = main.getApp ().getMainWindow ().getRemoteWindow ()
		for file in listing:
			entry = remote_window.parseListEntry (file)

			# Skip directory entries if the server provides them
			if entry[0] in ( '.', '..' ):
				continue

			# Copy over the truesize
			new_event = events.wxQueueEvent (
				file        = entry[0],
				host        = self.event.host,
				port        = self.event.port,
				username    = self.event.username,
				password    = self.event.password,
				size        = entry[1],
				flags       = entry[3][0],
				remote_path = self.event.remote_path,
				local_path  = self.event.local_path,
				direction   = self.event.direction,
				attempt     = 1,
				maxattempts = self.event.maxattempts,
				method      = self.event.method,
			)
			main.getApp ().postEvent (new_event)
